{-# LANGUAGE OverloadedStrings #-}

module PUBMED where

import Prelude hiding (takeWhile)
import Data.Text (Text)
import PUBMED.Client
import PUBMED.Parser
import Network.HTTP.Client (newManager)
import Network.HTTP.Client.TLS (tlsManagerSettings)
import Servant.Client (runClientM, mkClientEnv, BaseUrl(..), Scheme(..))
import Text.XML (parseLBS_, def)
import Text.XML.Cursor (fromDocument, Cursor)
import Text.XML.Stream.Parse (XmlException)
import Data.Conduit (ConduitT)
import Data.ByteString.Lazy (ByteString)
import Data.ByteString.Char8 (pack)
import Control.Monad.Catch (catch, MonadThrow, Exception)

import Control.Applicative
import Data.Attoparsec.ByteString
import Data.Attoparsec.ByteString.Char8 (anyChar)

import qualified Data.ByteString.Lazy as LBS
import qualified Data.ByteString      as DB
import qualified Data.Text as T

-- | API main function
getMetadataWith :: Text -> Maybe Limit -> IO (Either Text [PubMed])
getMetadataWith = runSimpleFindPubmedAbstractRequest

runParser :: Show res => (Cursor -> res) -> LBS.ByteString -> res
runParser parser = parser . fromDocument . parseLBS_ def

-- | TODO this parser need at least one subs at the end
-- (use endOfInput)
removeSub :: Parser ByteString
removeSub = do
  dt <- many textWithBalise
  eo <- manyTill anyChar endOfInput
  pure $ LBS.fromStrict $ pack $ concat dt <> eo
  where
    textWithBalise =
      manyTill anyChar (sub <|> asub)
    sub  = string "<sub>"
      <|> string "<sup>"
      <|> string "<i>"
      <|> string "<b>"
    asub = string "</sub>"
      <|> string "</sup>"
      <|> string "</i>"
      <|> string "</b>"

type Query = Text
type Limit = Integer

runMultipleFPAR ::
  [Integer]
  -> IO (Either Text [PubMed])
runMultipleFPAR ids
  | length ids < 300 = runSimpleFetchPubmedAbstractRequest ids
  | otherwise = do
      runSimpleFetchPubmedAbstractRequest (Prelude.take 300 ids)
        <> runMultipleFPAR (drop 300 ids)

runSimpleFetchPubmedAbstractRequest ::
  [Integer]
  -> IO (Either Text [PubMed])
runSimpleFetchPubmedAbstractRequest ids = do
  manager' <- newManager tlsManagerSettings
  res <- runClientM
    (fetch (Just "pubmed") (Just "abstract") ids)
    (mkClientEnv manager' $ BaseUrl Https "eutils.ncbi.nlm.nih.gov" 443 "entrez/eutils")
  case res of
    (Left err) -> pure (Left . T.pack $ show err)
    (Right (BsXml abs)) ->
      case parseOnly removeSub $ LBS.toStrict abs of
        (Left err'') -> pure (Left $ T.pack err'')
        (Right v) -> do
          parsed <- catch (pubMedParser v) ((\e -> do
                                                _ <- print e
                                                pure []) :: XmlException -> IO [PubMed])
          Right <$> pure parsed

runSimpleFindPubmedAbstractRequest :: Text -> Maybe Limit -> IO (Either Text [PubMed])
runSimpleFindPubmedAbstractRequest query limit =  do
  manager' <- newManager tlsManagerSettings
  res <- runClientM
    (search (Just query) limit)
    (mkClientEnv manager' $ BaseUrl Https "eutils.ncbi.nlm.nih.gov" 443 "entrez/eutils")
  case res of
    (Left err) -> pure (Left $ T.pack $ show err)
    (Right (BsXml docs)) -> do
      let docIds = runParser parseDocId docs
      runMultipleFPAR docIds

